from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Annotated

from fastapi import Body, FastAPI, HTTPException, Query
from fastapi.responses import JSONResponse

from lite_dist2.curriculum_models.curriculum import CurriculumProvider
from lite_dist2.curriculum_models.study import Study
from lite_dist2.curriculum_models.study_status import StudyStatus
from lite_dist2.curriculum_models.trial import Trial
from lite_dist2.expections import LD2ParameterError
from lite_dist2.table_node_api.table_param import StudyRegisterParam, TrialRegisterParam, TrialReserveParam
from lite_dist2.table_node_api.table_response import (
    CurriculumSummaryResponse,
    OkResponse,
    StudyRegisteredResponse,
    StudyResponse,
    TrialReserveResponse,
)

if TYPE_CHECKING:
    from pydantic import BaseModel

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

app = FastAPI()


def _response(model: BaseModel, status_code: int) -> JSONResponse:
    return JSONResponse(content=model.model_dump(mode="json"), status_code=status_code)


@app.get("/ping")
def handle_ping() -> OkResponse:
    return OkResponse(ok=True)


@app.get("/save")
def handle_save() -> OkResponse:
    curr = CurriculumProvider.get()
    curr.save()
    return OkResponse(ok=True)


@app.get("/status", response_model=CurriculumSummaryResponse)
def handle_status() -> CurriculumSummaryResponse | JSONResponse:
    curr = CurriculumProvider.get()
    return _response(CurriculumSummaryResponse(summaries=curr.to_summaries()), 200)


@app.post("/study/register", response_model=StudyRegisteredResponse)
def handle_study_register(
    study_registry: Annotated[StudyRegisterParam, Body(description="Registry of processing study")],
) -> StudyRegisteredResponse | JSONResponse:
    if not study_registry.study.is_valid():
        raise HTTPException(status_code=400, detail="Cannot use together infinite space and all_calculation strategy.")

    curr = CurriculumProvider.get()
    new_study = Study.from_model(study_registry.study.to_study_model())

    if curr.try_insert_study(new_study):
        return _response(StudyRegisteredResponse(study_id=new_study.study_id), 200)
    raise HTTPException(status_code=400, detail=f'The name("{new_study.name}") of study is already registered.')


@app.post("/trial/reserve", response_model=TrialReserveResponse)
def handle_trial_reserve(
    param: Annotated[TrialReserveParam, Body(description="Reserved trial parameter")],
) -> TrialReserveResponse | JSONResponse:
    curr = CurriculumProvider.get()
    study = curr.get_available_study(param.retaining_capacity)
    if study is None:
        return _response(TrialReserveResponse(trial=None), 202)

    trial = study.suggest_next_trial(param.max_size, param.worker_node_name, param.worker_node_id)
    if trial is None:
        return _response(TrialReserveResponse(trial=None), 202)
    return _response(TrialReserveResponse(trial=trial.to_model()), 200)


@app.post("/trial/register", response_model=OkResponse)
def handle_trial_register(
    param: Annotated[TrialRegisterParam, Body(description="Registering trial")],
) -> OkResponse | JSONResponse:
    curr = CurriculumProvider.get()
    trial = param.trial
    study = curr.find_study_by_id(trial.study_id)
    if study is None:
        raise HTTPException(status_code=404, detail=f"Study not found: study_id={trial.study_id}")

    try:
        study.receipt_trial(Trial.from_model(trial))
    except LD2ParameterError:
        return _response(OkResponse(ok=False), 400)
    curr.to_storage_if_done()
    return _response(OkResponse(ok=True), 200)


@app.get("/study", response_model=StudyResponse)
def handle_study(
    study_id: Annotated[str | None, Query(description="`study_id` of the target study")] = None,
    name: Annotated[str | None, Query(description="`name` of the target study")] = None,
) -> StudyResponse | JSONResponse:
    if study_id is None and name is None:
        raise HTTPException(status_code=400, detail="One of study_id or name should be set.")
    if study_id is not None and name is not None:
        raise HTTPException(status_code=400, detail="Only one of study_id or name should be set.")

    curr = CurriculumProvider.get()
    storage = curr.pop_storage(study_id, name)
    if storage is not None:
        return _response(StudyResponse(status=StudyStatus.done, result=storage), 200)

    study_status = curr.get_study_status(study_id, name)
    resp = StudyResponse(status=study_status, result=None)
    if study_status == StudyStatus.not_found:
        return _response(resp, status_code=404)
    return _response(resp, 202)


@app.delete("/study", response_model=OkResponse)
def handle_study_cancel(
    study_id: Annotated[str | None, Query(description="`study_id` of the target study")] = None,
    name: Annotated[str | None, Query(description="`name` of the target study")] = None,
) -> OkResponse | JSONResponse:
    if study_id is None and name is None:
        raise HTTPException(status_code=400, detail="One of study_id or name should be set.")
    if study_id is not None and name is not None:
        raise HTTPException(status_code=400, detail="Only one of study_id or name should be set.")

    curr = CurriculumProvider.get()
    try:
        found_and_cancel = curr.cancel_study(study_id, name)
    except LD2ParameterError:
        return _response(OkResponse(ok=False), 400)

    if found_and_cancel:
        return _response(OkResponse(ok=True), 200)
    return _response(OkResponse(ok=False), 404)



from datetime import timedelta

import pytest

from lite_dist2.curriculum_models.trial import Mapping, Trial, TrialModel, TrialStatus
from lite_dist2.curriculum_models.trial_table import TrialTable, TrialTableModel
from lite_dist2.expections import LD2ParameterError
from lite_dist2.value_models.aligned_space import ParameterAlignedSpace, ParameterAlignedSpaceModel
from lite_dist2.value_models.base_space import FlattenSegment
from lite_dist2.value_models.jagged_space import ParameterJaggedSpace
from lite_dist2.value_models.line_segment import DummyLineSegment, LineSegmentModel, ParameterRangeInt
from lite_dist2.value_models.point import ScalarValue
from tests.const import DT

_DUMMY_PARAMETER_SPACE = ParameterAlignedSpace(
    axes=[
        ParameterRangeInt(name="x", type="int", size=2, start=0, ambient_size=2, ambient_index=0),
        ParameterRangeInt(name="y", type="int", size=2, start=0, ambient_size=2, ambient_index=0),
    ],
    check_lower_filling=True,
)


@pytest.mark.parametrize(
    ("table", "total_num", "expected"),
    [
        pytest.param(
            TrialTable(
                trials=[],  # ここでは使わないので空
                aggregated_parameter_space=None,
            ),
            100,
            FlattenSegment(0, None),
            id="Empty (None aps)",
        ),
        pytest.param(
            TrialTable(
                trials=[],
                aggregated_parameter_space={-1: [], 0: []},
            ),
            100,
            FlattenSegment(0, None),
            id="Empty",
        ),
        pytest.param(
            TrialTable(
                trials=[],
                aggregated_parameter_space={
                    -1: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=10,
                                    ambient_index=0,
                                    ambient_size=10,
                                    start=0,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                    0: [],
                },
            ),
            10,
            FlattenSegment(10, 0),
            id="Universal 1D",
        ),
        pytest.param(
            TrialTable(
                trials=[],
                aggregated_parameter_space={
                    -1: [],
                    0: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=10,
                                    ambient_index=0,
                                    ambient_size=None,
                                    start=0,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            None,
            FlattenSegment(10, None),
            id="Infinite 1D",
        ),
        pytest.param(
            TrialTable(
                trials=[],
                aggregated_parameter_space={
                    -1: [],
                    0: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=10,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=10,
                                    ambient_index=50,
                                    ambient_size=100,
                                    start=50,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            10,
            FlattenSegment(10, 40),
            id="Segmented 1D",
        ),
        pytest.param(
            TrialTable(
                trials=[],
                aggregated_parameter_space={
                    -1: [],
                    0: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=10,
                                    ambient_index=0,
                                    ambient_size=None,
                                    start=0,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=10,
                                    ambient_index=50,
                                    ambient_size=None,
                                    start=50,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            None,
            FlattenSegment(10, 40),
            id="Infinite segmented 1D",
        ),
        pytest.param(
            TrialTable(
                trials=[],
                aggregated_parameter_space={
                    -1: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=10,
                                    ambient_index=0,
                                    ambient_size=10,
                                    start=0,
                                    step=1,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=10,
                                    ambient_index=0,
                                    ambient_size=10,
                                    start=0,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                    0: [],
                    1: [],
                },
            ),
            100,
            FlattenSegment(100, 0),
            id="Universal 2D",
        ),
        pytest.param(
            TrialTable(
                trials=[],
                aggregated_parameter_space={
                    -1: [],
                    0: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    ambient_index=1,
                                    ambient_size=10,
                                    start=1,
                                    step=1,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=10,
                                    ambient_index=0,
                                    ambient_size=10,
                                    start=0,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                    1: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=10,
                                    start=0,
                                    step=1,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=5,
                                    ambient_index=0,
                                    ambient_size=10,
                                    start=0,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            100,
            FlattenSegment(5, 5),
            id="Segmented 2D",
        ),
        pytest.param(
            TrialTable(
                trials=[
                    Trial(
                        study_id="s01",
                        trial_id="t01",
                        timestamp=DT,
                        parameter_space=ParameterJaggedSpace(
                            parameters=[
                                (1, 0),
                                (1, 1),
                                (1, 2),
                                (1, 3),
                                (1, 4),
                                (1, 5),
                                (1, 6),
                                (1, 7),
                                (1, 8),
                                (1, 9),
                            ],
                            ambient_indices=[
                                (1, 0),
                                (1, 1),
                                (1, 2),
                                (1, 3),
                                (1, 4),
                                (1, 5),
                                (1, 6),
                                (1, 7),
                                (1, 8),
                                (1, 9),
                            ],
                            axes_info=[
                                DummyLineSegment(name="x", type="int", ambient_size=10, step=1),
                                DummyLineSegment(name="y", type="int", ambient_size=10, step=1),
                            ],
                        ),
                        trial_status=TrialStatus.done,
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0xa", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0xb", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x2", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0xc", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x3", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0xd", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x4", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0xe", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x5", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0xf", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x6", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x10", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x7", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x11", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x8", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x12", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x9", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x13", name="p"),
                            ),
                        ],
                    ),
                    Trial(
                        study_id="s01",
                        trial_id="t02",
                        timestamp=DT,
                        parameter_space=ParameterJaggedSpace(
                            parameters=[
                                (0, 0),
                                (0, 1),
                                (0, 2),
                                (0, 3),
                                (0, 4),
                            ],
                            ambient_indices=[
                                (0, 0),
                                (0, 1),
                                (0, 2),
                                (0, 3),
                                (0, 4),
                            ],
                            axes_info=[
                                DummyLineSegment(name="x", type="int", ambient_size=10, step=1),
                                DummyLineSegment(name="y", type="int", ambient_size=10, step=1),
                            ],
                        ),
                        trial_status=TrialStatus.done,
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x0", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x1", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x2", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x2", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x3", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x3", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x4", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x4", name="p"),
                            ),
                        ],
                    ),
                    Trial(
                        study_id="s01",
                        trial_id="t03",
                        timestamp=DT,
                        parameter_space=ParameterJaggedSpace(
                            parameters=[
                                (0, 5),
                                (0, 6),
                                (0, 7),
                                (0, 8),
                                (0, 9),
                            ],
                            ambient_indices=[
                                (0, 5),
                                (0, 6),
                                (0, 7),
                                (0, 8),
                                (0, 9),
                            ],
                            axes_info=[
                                DummyLineSegment(name="x", type="int", ambient_size=10, step=1),
                                DummyLineSegment(name="y", type="int", ambient_size=10, step=1),
                            ],
                        ),
                        trial_status=TrialStatus.running,
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                    ),
                ],
                aggregated_parameter_space={
                    -1: [],
                    0: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    ambient_index=1,
                                    ambient_size=10,
                                    start=1,
                                    step=1,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=10,
                                    ambient_index=0,
                                    ambient_size=10,
                                    start=0,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                    1: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=10,
                                    start=0,
                                    step=1,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=5,
                                    ambient_index=0,
                                    ambient_size=10,
                                    start=0,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            100,
            FlattenSegment(20, None),
            id="Infinite jagged running 2D",
        ),
        pytest.param(
            TrialTable(
                trials=[
                    Trial(
                        study_id="s01",
                        trial_id="t01",
                        timestamp=DT,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    step=1,
                                    size=1,
                                    start=1,
                                    ambient_index=1,
                                    ambient_size=10,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    step=1,
                                    size=10,
                                    start=0,
                                    ambient_index=0,
                                    ambient_size=10,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        trial_status=TrialStatus.done,
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0xa", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0xb", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x2", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0xc", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x3", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0xd", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x4", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0xe", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x5", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0xf", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x6", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x10", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x7", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x11", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x8", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x12", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x9", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x13", name="p"),
                            ),
                        ],
                    ),
                    Trial(
                        study_id="s01",
                        trial_id="t02",
                        timestamp=DT,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    step=1,
                                    size=1,
                                    start=0,
                                    ambient_index=0,
                                    ambient_size=10,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    step=1,
                                    size=5,
                                    start=0,
                                    ambient_index=0,
                                    ambient_size=10,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        trial_status=TrialStatus.done,
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x0", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x1", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x1", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x2", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x2", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x3", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x3", name="p"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x4", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x4", name="p"),
                            ),
                        ],
                    ),
                    Trial(
                        study_id="s01",
                        trial_id="t03",
                        timestamp=DT,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    step=1,
                                    size=1,
                                    start=0,
                                    ambient_index=0,
                                    ambient_size=10,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    step=1,
                                    size=5,
                                    start=5,
                                    ambient_index=5,
                                    ambient_size=10,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        trial_status=TrialStatus.running,
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                    ),
                ],
                aggregated_parameter_space={
                    -1: [],
                    0: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    ambient_index=1,
                                    ambient_size=10,
                                    start=1,
                                    step=1,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=10,
                                    ambient_index=0,
                                    ambient_size=10,
                                    start=0,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                    1: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=10,
                                    start=0,
                                    step=1,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=5,
                                    ambient_index=0,
                                    ambient_size=10,
                                    start=0,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            100,
            FlattenSegment(20, None),
            id="Infinite aligned running 2D",
        ),
    ],
)
def test_trial_table_find_least_division(
    table: TrialTable,
    total_num: int,
    expected: FlattenSegment,
) -> None:
    actual = table.find_least_division(total_num)
    assert actual == expected


@pytest.mark.parametrize(
    ("table", "expected"),
    [
        pytest.param(
            TrialTable(
                trials=[],  # ここでは使わないので空
                aggregated_parameter_space=None,
            ),
            0,
            id="Empty_aps=None",
        ),
        pytest.param(
            TrialTable(
                trials=[],
                aggregated_parameter_space={-1: [], 0: []},
            ),
            0,
            id="Empty",
        ),
        pytest.param(
            TrialTable(
                trials=[],
                aggregated_parameter_space={
                    -1: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=10,
                                    ambient_index=0,
                                    ambient_size=10,
                                    start=0,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                    0: [],
                },
            ),
            10,
            id="Universal 1D",
        ),
        pytest.param(
            TrialTable(
                trials=[],
                aggregated_parameter_space={
                    -1: [],
                    0: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=10,
                                    ambient_index=0,
                                    ambient_size=None,
                                    start=0,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            10,
            id="Infinite 1D",
        ),
        pytest.param(
            TrialTable(
                trials=[],
                aggregated_parameter_space={
                    -1: [],
                    0: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=10,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=10,
                                    ambient_index=50,
                                    ambient_size=100,
                                    start=50,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            20,
            id="Segmented 1D",
        ),
        pytest.param(
            TrialTable(
                trials=[],
                aggregated_parameter_space={
                    -1: [],
                    0: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=10,
                                    ambient_index=0,
                                    ambient_size=None,
                                    start=0,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=10,
                                    ambient_index=50,
                                    ambient_size=None,
                                    start=50,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            20,
            id="Infinite segmented 1D",
        ),
        pytest.param(
            TrialTable(
                trials=[],
                aggregated_parameter_space={
                    -1: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=10,
                                    ambient_index=0,
                                    ambient_size=10,
                                    start=0,
                                    step=1,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=10,
                                    ambient_index=0,
                                    ambient_size=10,
                                    start=0,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                    0: [],
                    1: [],
                },
            ),
            100,
            id="Universal 2D",
        ),
        pytest.param(
            TrialTable(
                trials=[],
                aggregated_parameter_space={
                    -1: [],
                    0: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    ambient_index=1,
                                    ambient_size=10,
                                    start=1,
                                    step=1,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=10,
                                    ambient_index=0,
                                    ambient_size=10,
                                    start=0,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                    1: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    name="x",
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=10,
                                    start=0,
                                    step=1,
                                ),
                                ParameterRangeInt(
                                    name="y",
                                    type="int",
                                    size=5,
                                    ambient_index=0,
                                    ambient_size=10,
                                    start=0,
                                    step=1,
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            15,
            id="Segmented 2D",
        ),
    ],
)
def test_trial_table_count_grid(
    table: TrialTable,
    expected: int,
) -> None:
    actual = table.count_grid()
    assert actual == expected


@pytest.mark.parametrize(
    "model",
    [
        TrialTableModel(
            trials=[],
            aggregated_parameter_space=None,
        ),
        TrialTableModel(
            trials=[
                TrialModel(
                    study_id="some_study",
                    trial_id="01",
                    timestamp=DT,
                    trial_status=TrialStatus.running,
                    parameter_space=ParameterAlignedSpaceModel(
                        type="aligned",
                        axes=[
                            LineSegmentModel(
                                type="int",
                                size="0xa",
                                step="0x1",
                                start="0x0",
                                ambient_index="0x0",
                                ambient_size="0x64",
                            ),
                        ],
                        check_lower_filling=True,
                    ),
                    result_type="scalar",
                    result_value_type="float",
                    worker_node_name="w01",
                    worker_node_id="w01",
                ),
            ],
            aggregated_parameter_space=None,
        ),
        TrialTableModel(
            trials=[
                TrialModel(
                    study_id="some_study",
                    trial_id="01",
                    timestamp=DT,
                    trial_status=TrialStatus.running,
                    parameter_space=ParameterAlignedSpaceModel(
                        type="aligned",
                        axes=[
                            LineSegmentModel(
                                type="int",
                                size="0xa",
                                step="0x1",
                                start="0x0",
                                ambient_index="0x0",
                                ambient_size="0x64",
                            ),
                        ],
                        check_lower_filling=True,
                    ),
                    result_type="scalar",
                    result_value_type="float",
                    worker_node_name="w01",
                    worker_node_id="w01",
                    results=[
                        Mapping(
                            params=(
                                ScalarValue(
                                    type="scalar",
                                    value_type="int",
                                    value="0x1",
                                    name="p1",
                                ),
                            ),
                            result=ScalarValue(
                                type="scalar",
                                value_type="float",
                                value="0x1.0000000000000p+1",
                                name="p2",
                            ),
                        ),
                    ],
                ),
            ],
            aggregated_parameter_space=None,
        ),
    ],
)
def test_trial_table_to_model_from_model(model: TrialTableModel) -> None:
    table = TrialTable.from_model(model)
    reconstructed_model = table.to_model()
    assert model == reconstructed_model


@pytest.mark.parametrize(
    ("trial_table", "trial_id", "worker_node_id", "result", "expected"),
    [
        pytest.param(
            TrialTable(
                trials=[
                    Trial(
                        study_id="s01",
                        trial_id="t01",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x0",
                                        name="x",
                                    ),
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x0",
                                        name="y",
                                    ),
                                ),
                                result=ScalarValue(
                                    type="scalar",
                                    value_type="int",
                                    value="0x0",
                                    name="p2",
                                ),
                            ),
                        ],
                    ),
                    Trial(
                        study_id="s01",
                        trial_id="t02",
                        timestamp=DT,
                        trial_status=TrialStatus.running,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=1,
                                    ambient_size=100,
                                    start=1,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=None,
                    ),
                    Trial(
                        study_id="s01",
                        trial_id="t03",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=2,
                                    ambient_size=100,
                                    start=2,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=2,
                                    ambient_size=100,
                                    start=2,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x0",
                                        name="x",
                                    ),
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x2",
                                        name="y",
                                    ),
                                ),
                                result=ScalarValue(
                                    type="scalar",
                                    value_type="int",
                                    value="0x2",
                                    name="p2",
                                ),
                            ),
                        ],
                    ),
                ],
                aggregated_parameter_space={
                    -1: [],
                    0: [],
                    1: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=2,
                                    ambient_size=100,
                                    start=2,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            "t02",
            "w01",
            None,
            TrialTable(
                trials=[
                    Trial(
                        study_id="s01",
                        trial_id="t01",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x0",
                                        name="x",
                                    ),
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x0",
                                        name="y",
                                    ),
                                ),
                                result=ScalarValue(
                                    type="scalar",
                                    value_type="int",
                                    value="0x0",
                                    name="p2",
                                ),
                            ),
                        ],
                    ),
                    Trial(
                        study_id="s01",
                        trial_id="t02",
                        timestamp=DT,
                        trial_status=TrialStatus.running,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=1,
                                    ambient_size=100,
                                    start=1,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=None,
                    ),
                    Trial(
                        study_id="s01",
                        trial_id="t03",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=2,
                                    ambient_size=100,
                                    start=2,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=2,
                                    ambient_size=100,
                                    start=2,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x0",
                                        name="x",
                                    ),
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x2",
                                        name="y",
                                    ),
                                ),
                                result=ScalarValue(
                                    type="scalar",
                                    value_type="int",
                                    value="0x2",
                                    name="p2",
                                ),
                            ),
                        ],
                    ),
                ],
                aggregated_parameter_space={
                    -1: [],
                    0: [],
                    1: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=2,
                                    ambient_size=100,
                                    start=2,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            id="Do nothing: None result",
        ),
        pytest.param(
            TrialTable(
                trials=[
                    Trial(
                        study_id="s01",
                        trial_id="t01",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x0",
                                        name="x",
                                    ),
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x0",
                                        name="y",
                                    ),
                                ),
                                result=ScalarValue(
                                    type="scalar",
                                    value_type="int",
                                    value="0x0",
                                    name="p2",
                                ),
                            ),
                        ],
                    ),
                    Trial(
                        study_id="s01",
                        trial_id="t02",
                        timestamp=DT,
                        trial_status=TrialStatus.running,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=1,
                                    ambient_size=100,
                                    start=1,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=None,
                    ),
                    Trial(
                        study_id="s01",
                        trial_id="t03",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=2,
                                    ambient_size=100,
                                    start=2,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=2,
                                    ambient_size=100,
                                    start=2,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x0",
                                        name="x",
                                    ),
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x2",
                                        name="y",
                                    ),
                                ),
                                result=ScalarValue(
                                    type="scalar",
                                    value_type="int",
                                    value="0x2",
                                    name="p2",
                                ),
                            ),
                        ],
                    ),
                ],
                aggregated_parameter_space={
                    -1: [],
                    0: [],
                    1: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=2,
                                    ambient_size=100,
                                    start=2,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            "t02",
            "w01",
            [
                Mapping(
                    params=(
                        ScalarValue(
                            type="scalar",
                            value_type="int",
                            value="0x0",
                            name="x",
                        ),
                        ScalarValue(
                            type="scalar",
                            value_type="int",
                            value="0x1",
                            name="y",
                        ),
                    ),
                    result=ScalarValue(
                        type="scalar",
                        value_type="int",
                        value="0x1",
                        name="p2",
                    ),
                ),
            ],
            TrialTable(
                trials=[
                    Trial(
                        study_id="s01",
                        trial_id="t01",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x0",
                                        name="x",
                                    ),
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x0",
                                        name="y",
                                    ),
                                ),
                                result=ScalarValue(
                                    type="scalar",
                                    value_type="int",
                                    value="0x0",
                                    name="p2",
                                ),
                            ),
                        ],
                    ),
                    Trial(
                        study_id="s01",
                        trial_id="t02",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=1,
                                    ambient_size=100,
                                    start=1,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x0",
                                        name="x",
                                    ),
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x1",
                                        name="y",
                                    ),
                                ),
                                result=ScalarValue(
                                    type="scalar",
                                    value_type="int",
                                    value="0x1",
                                    name="p2",
                                ),
                            ),
                        ],
                    ),
                    Trial(
                        study_id="s01",
                        trial_id="t03",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=2,
                                    ambient_size=100,
                                    start=2,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=2,
                                    ambient_size=100,
                                    start=2,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x0",
                                        name="x",
                                    ),
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x2",
                                        name="y",
                                    ),
                                ),
                                result=ScalarValue(
                                    type="scalar",
                                    value_type="int",
                                    value="0x2",
                                    name="p2",
                                ),
                            ),
                        ],
                    ),
                ],
                aggregated_parameter_space={
                    -1: [],
                    0: [],
                    1: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=2,
                                    ambient_size=100,
                                    start=2,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=100,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=1,
                                    ambient_size=100,
                                    start=1,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            id="aligned normal",
        ),
        pytest.param(
            TrialTable(
                trials=[
                    Trial(
                        study_id="s01",
                        trial_id="t01",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterJaggedSpace(
                            parameters=[(0, 0)],
                            ambient_indices=[(0, 0)],
                            axes_info=[
                                DummyLineSegment(type="int", name="x", ambient_size=6, step=1),
                                DummyLineSegment(type="int", name="y", ambient_size=6, step=1),
                            ],
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x0",
                                        name="x",
                                    ),
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x0",
                                        name="y",
                                    ),
                                ),
                                result=ScalarValue(
                                    type="scalar",
                                    value_type="int",
                                    value="0x0",
                                    name="p2",
                                ),
                            ),
                        ],
                    ),
                    Trial(
                        study_id="s01",
                        trial_id="t02",
                        timestamp=DT,
                        trial_status=TrialStatus.running,
                        parameter_space=ParameterJaggedSpace(
                            parameters=[(0, 1)],
                            ambient_indices=[(0, 1)],
                            axes_info=[
                                DummyLineSegment(type="int", name="x", ambient_size=6, step=1),
                                DummyLineSegment(type="int", name="y", ambient_size=6, step=1),
                            ],
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=None,
                    ),
                    Trial(
                        study_id="s01",
                        trial_id="t03",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterJaggedSpace(
                            parameters=[(0, 2)],
                            ambient_indices=[(0, 2)],
                            axes_info=[
                                DummyLineSegment(type="int", name="x", ambient_size=6, step=1),
                                DummyLineSegment(type="int", name="y", ambient_size=6, step=1),
                            ],
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x0",
                                        name="x",
                                    ),
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x2",
                                        name="y",
                                    ),
                                ),
                                result=ScalarValue(
                                    type="scalar",
                                    value_type="int",
                                    value="0x2",
                                    name="p2",
                                ),
                            ),
                        ],
                    ),
                ],
                aggregated_parameter_space={
                    -1: [],
                    0: [],
                    1: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=6,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=6,
                                    start=0,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=6,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=2,
                                    ambient_size=6,
                                    start=2,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            "t02",
            "w01",
            [
                Mapping(
                    params=(
                        ScalarValue(
                            type="scalar",
                            value_type="int",
                            value="0x0",
                            name="x",
                        ),
                        ScalarValue(
                            type="scalar",
                            value_type="int",
                            value="0x1",
                            name="y",
                        ),
                    ),
                    result=ScalarValue(
                        type="scalar",
                        value_type="int",
                        value="0x1",
                        name="p2",
                    ),
                ),
            ],
            TrialTable(
                trials=[
                    Trial(
                        study_id="s01",
                        trial_id="t01",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterJaggedSpace(
                            parameters=[(0, 0)],
                            ambient_indices=[(0, 0)],
                            axes_info=[
                                DummyLineSegment(type="int", name="x", ambient_size=6, step=1),
                                DummyLineSegment(type="int", name="y", ambient_size=6, step=1),
                            ],
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x0",
                                        name="x",
                                    ),
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x0",
                                        name="y",
                                    ),
                                ),
                                result=ScalarValue(
                                    type="scalar",
                                    value_type="int",
                                    value="0x0",
                                    name="p2",
                                ),
                            ),
                        ],
                    ),
                    Trial(
                        study_id="s01",
                        trial_id="t02",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterJaggedSpace(
                            parameters=[(0, 1)],
                            ambient_indices=[(0, 1)],
                            axes_info=[
                                DummyLineSegment(type="int", name="x", ambient_size=6, step=1),
                                DummyLineSegment(type="int", name="y", ambient_size=6, step=1),
                            ],
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x0",
                                        name="x",
                                    ),
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x1",
                                        name="y",
                                    ),
                                ),
                                result=ScalarValue(
                                    type="scalar",
                                    value_type="int",
                                    value="0x1",
                                    name="p2",
                                ),
                            ),
                        ],
                    ),
                    Trial(
                        study_id="s01",
                        trial_id="t03",
                        timestamp=DT,
                        trial_status=TrialStatus.done,
                        parameter_space=ParameterJaggedSpace(
                            parameters=[(0, 2)],
                            ambient_indices=[(0, 2)],
                            axes_info=[
                                DummyLineSegment(type="int", name="x", ambient_size=6, step=1),
                                DummyLineSegment(type="int", name="y", ambient_size=6, step=1),
                            ],
                        ),
                        result_type="scalar",
                        result_value_type="int",
                        worker_node_name="w01",
                        worker_node_id="w01",
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x0",
                                        name="x",
                                    ),
                                    ScalarValue(
                                        type="scalar",
                                        value_type="int",
                                        value="0x2",
                                        name="y",
                                    ),
                                ),
                                result=ScalarValue(
                                    type="scalar",
                                    value_type="int",
                                    value="0x2",
                                    name="p2",
                                ),
                            ),
                        ],
                    ),
                ],
                aggregated_parameter_space={
                    -1: [],
                    0: [],
                    1: [
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=6,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=6,
                                    start=0,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=6,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=2,
                                    ambient_size=6,
                                    start=2,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                        ParameterAlignedSpace(
                            axes=[
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=0,
                                    ambient_size=6,
                                    start=0,
                                    name="x",
                                ),
                                ParameterRangeInt(
                                    type="int",
                                    size=1,
                                    ambient_index=1,
                                    ambient_size=6,
                                    start=1,
                                    name="y",
                                ),
                            ],
                            check_lower_filling=True,
                        ),
                    ],
                },
            ),
            id="jagged normal",
        ),
    ],
)
def test_trial_table_receipt_trial_result(
    trial_table: TrialTable,
    trial_id: str,
    worker_node_id: str,
    result: list[Mapping] | None,
    expected: TrialTable,
) -> None:
    trial_table.receipt_trial_result(trial_id, result, worker_node_id)
    actual_model = trial_table.to_model()
    expected_model = expected.to_model()
    assert actual_model.trials == expected_model.trials
    assert actual_model.aggregated_parameter_space == expected_model.aggregated_parameter_space


# noinspection SpellCheckingInspection
def test_trial_table_receipt_trial_result_raise_override_done_trial() -> None:
    trial_table = TrialTable(
        trials=[
            Trial(
                study_id="s01",
                trial_id="t01",
                timestamp=DT,
                trial_status=TrialStatus.done,
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=1, ambient_index=0, ambient_size=100, start=0, name="x"),
                        ParameterRangeInt(type="int", size=1, ambient_index=0, ambient_size=100, start=0, name="y"),
                    ],
                    check_lower_filling=True,
                ),
                result_type="scalar",
                result_value_type="int",
                worker_node_name="w01",
                worker_node_id="w01",
                results=[
                    Mapping(
                        params=(
                            ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                            ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                        ),
                        result=ScalarValue(type="scalar", value_type="int", value="0x0", name="p2"),
                    ),
                ],
            ),
        ],
        aggregated_parameter_space={
            -1: [],
            0: [],
            1: [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=1, ambient_index=0, ambient_size=100, start=0, name="x"),
                        ParameterRangeInt(type="int", size=1, ambient_index=0, ambient_size=100, start=0, name="y"),
                    ],
                    check_lower_filling=True,
                ),
            ],
        },
    )

    result = [
        Mapping(
            params=(
                ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
            ),
            result=ScalarValue(type="scalar", value_type="int", value="0x0", name="p2"),
        ),
    ]

    with pytest.raises(LD2ParameterError, match=r"Cannot\soverride\sresult\sof\sdone\strial"):
        trial_table.receipt_trial_result("t01", result, "w01")


# noinspection SpellCheckingInspection
def test_trial_table_receipt_trial_result_raise_not_found_trial() -> None:
    trial_table = TrialTable(
        trials=[],
        aggregated_parameter_space={
            -1: [],
            0: [],
            1: [],
        },
    )

    result = [
        Mapping(
            params=(
                ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
            ),
            result=ScalarValue(type="scalar", value_type="int", value="0x0", name="p2"),
        ),
    ]

    with pytest.raises(LD2ParameterError, match=r"Not\sfound\strial\sthat\sid"):
        trial_table.receipt_trial_result("t01", result, "w01")


# noinspection SpellCheckingInspection
def test_trial_table_receipt_trial_result_raise_unmatch_worker_id() -> None:
    trial_table = TrialTable(
        trials=[
            Trial(
                study_id="s01",
                trial_id="t01",
                timestamp=DT,
                trial_status=TrialStatus.running,
                parameter_space=ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=1, ambient_index=0, ambient_size=100, start=0, name="x"),
                        ParameterRangeInt(type="int", size=1, ambient_index=0, ambient_size=100, start=0, name="y"),
                    ],
                    check_lower_filling=True,
                ),
                result_type="scalar",
                result_value_type="int",
                worker_node_name="w01",
                worker_node_id="w01",
                results=None,
            ),
        ],
        aggregated_parameter_space={
            -1: [],
            0: [],
            1: [],
        },
    )

    result = [
        Mapping(
            params=(
                ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
            ),
            result=ScalarValue(type="scalar", value_type="int", value="0x0", name="p2"),
        ),
    ]

    with pytest.raises(LD2ParameterError, match=r"This\strial\sis\sreserved\sby\sother\sworker"):
        trial_table.receipt_trial_result("t01", result, "w02")


def test_trial_table_check_timeout_trial() -> None:
    common_param = {
        "study_id": "s01",
        "parameter_space": _DUMMY_PARAMETER_SPACE,
        "result_type": "scalar",
        "result_value_type": "int",
        "results": [],
        "worker_node_name": "w01",
        "worker_node_id": "w01",
    }

    now = DT
    trial_table = TrialTable(
        trials=[
            Trial(
                trial_id="running_but_created_now",
                timestamp=now,
                trial_status=TrialStatus.running,
                **common_param,
            ),
            Trial(
                trial_id="running_little_past",
                timestamp=now - timedelta(seconds=30),
                trial_status=TrialStatus.running,
                **common_param,
            ),
            Trial(
                trial_id="running_very_past",
                timestamp=now - timedelta(seconds=3000),
                trial_status=TrialStatus.running,
                **common_param,
            ),
            Trial(
                trial_id="done_little_past",
                timestamp=now - timedelta(seconds=40),
                trial_status=TrialStatus.done,
                **common_param,
            ),
            Trial(
                trial_id="done_very_past",
                timestamp=now - timedelta(seconds=3000),
                trial_status=TrialStatus.done,
                **common_param,
            ),
        ],
        aggregated_parameter_space={},
    )

    expected_trial_table = TrialTable(
        trials=[
            Trial(
                trial_id="running_but_created_now",
                timestamp=now,
                trial_status=TrialStatus.running,
                **common_param,
            ),
            Trial(
                trial_id="running_little_past",
                timestamp=now - timedelta(seconds=30),
                trial_status=TrialStatus.running,
                **common_param,
            ),
            Trial(
                trial_id="done_little_past",
                timestamp=now - timedelta(seconds=40),
                trial_status=TrialStatus.done,
                **common_param,
            ),
            Trial(
                trial_id="done_very_past",
                timestamp=now - timedelta(seconds=3000),
                trial_status=TrialStatus.done,
                **common_param,
            ),
        ],
        aggregated_parameter_space={},
    )
    expected_ids = ["running_very_past"]

    actual_ids = trial_table.check_timeout_trial(now, timeout_seconds=300)
    assert actual_ids == expected_ids
    assert trial_table.to_model() == expected_trial_table.to_model()



from typing import Literal

from pydantic import BaseModel

from lite_dist2.type_definitions import PortableValueType
from lite_dist2.value_models.aligned_space import ParameterAlignedSpaceModel
from lite_dist2.value_models.line_segment import LineSegmentModel


class LineSegmentRegistry(BaseModel):
    name: str | None = None
    type: Literal["bool", "int", "float"]
    size: str | None
    step: PortableValueType
    start: PortableValueType

    def to_line_segment_model(self) -> LineSegmentModel:
        return LineSegmentModel(
            name=self.name,
            type=self.type,
            size=self.size,
            step=self.step,
            start=self.start,
            ambient_index="0x0",
            ambient_size=self.size,
            is_dummy=False,
        )


class ParameterAlignedSpaceRegistry(BaseModel):
    type: Literal["aligned"]
    axes: list[LineSegmentRegistry]

    def to_parameter_aligned_space_model(self) -> ParameterAlignedSpaceModel:
        return ParameterAlignedSpaceModel(
            type=self.type,
            axes=[axis.to_line_segment_model() for axis in self.axes],
            check_lower_filling=True,
        )



from __future__ import annotations

import itertools
from typing import TYPE_CHECKING

from pydantic import BaseModel

from lite_dist2.curriculum_models.trial import Mapping, Trial, TrialModel, TrialStatus
from lite_dist2.expections import LD2ParameterError
from lite_dist2.value_models.aligned_space import ParameterAlignedSpace, ParameterAlignedSpaceModel
from lite_dist2.value_models.base_space import FlattenSegment
from lite_dist2.value_models.parameter_aligned_space_helper import remap_space, simplify

if TYPE_CHECKING:
    from datetime import datetime

    from lite_dist2.value_models.point import ResultType


class TrialTableModel(BaseModel):
    trials: list[TrialModel]
    aggregated_parameter_space: dict[int, list[ParameterAlignedSpaceModel]] | None

    @staticmethod
    def create_empty() -> TrialTableModel:
        return TrialTableModel(
            trials=[],
            aggregated_parameter_space=None,
        )


class TrialTable:
    def __init__(
        self,
        trials: list[Trial],
        aggregated_parameter_space: dict[int, list[ParameterAlignedSpace]] | None,
    ) -> None:
        self.trials = trials
        self.aggregated_parameter_space = aggregated_parameter_space

    def is_not_defined_aps(self) -> bool:
        return self.aggregated_parameter_space is None

    def is_empty(self) -> bool:
        return self.is_not_defined_aps() or len(self.trials) == 0

    def register(self, trial: Trial) -> None:
        self.trials.append(trial)

    def receipt_trial_result(self, receipted_trial_id: str, result: list[Mapping] | None, worker_node_id: str) -> None:
        if result is None:
            # Do nothing
            return

        for trial in reversed(self.trials):
            if trial.trial_id != receipted_trial_id:
                continue
            if trial.worker_node_id != worker_node_id:
                p = "worker_node_id"
                t = "This trial is reserved by other worker"
                raise LD2ParameterError(p, t)
            if trial.trial_status == TrialStatus.done:
                p = "receipted_trial_id"
                t = f"Cannot override result of done trial(id={receipted_trial_id})"
                raise LD2ParameterError(p, t)

            # Normal
            trial.result = result
            trial.trial_status = TrialStatus.done
            self.aggregated_parameter_space[self.trials[0].parameter_space.get_dim() - 1].extend(
                trial.parameter_space.to_aligned_list(),
            )
            return

        p = "receipted_trial_id"
        t = f"Not found trial that id={receipted_trial_id}"
        raise LD2ParameterError(p, t)

    def count_grid(self) -> int:
        if self.aggregated_parameter_space is None:
            return 0
        return sum(sum(space.total for space in spaces) for spaces in list(self.aggregated_parameter_space.values()))

    def count_trial(self) -> int:
        return len(self.trials)

    def simplify_aps(self) -> None:
        if self.aggregated_parameter_space is None:
            return

        dim = max(self.aggregated_parameter_space.keys()) + 1
        remapped_spaces: list[dict[int, list[ParameterAlignedSpace]]] = []
        for d in reversed(range(dim)):
            simplified = simplify(self.aggregated_parameter_space[d], d)
            remapped_spaces.append(remap_space(simplified, dim))

        self.aggregated_parameter_space = {
            d: list(itertools.chain.from_iterable(remapped_space[d] for remapped_space in remapped_spaces))
            for d in range(-1, dim)
        }

    def find_least_division(self, total_num: int | None) -> FlattenSegment:
        if self.aggregated_parameter_space is None:
            return FlattenSegment(0, None)

        aps_segments = [
            space.get_flatten_ambient_start_and_size()
            for spaces in self.aggregated_parameter_space.values()
            for space in spaces
        ]

        running_segments = simplify([segment for trial in self.trials for segment in trial.get_running_segments()])

        merged = simplify(aps_segments + running_segments)
        match len(merged):
            case 0:
                return FlattenSegment(0, None)
            case 1:
                start_index = merged[0].next_start_index()
                if total_num is None or start_index < total_num:
                    return FlattenSegment(start_index, None)
                return FlattenSegment(start_index, 0)
            case _:
                start = merged[0].next_start_index()
                return FlattenSegment(start, merged[1].get_start_index() - start)

    def init_aps(self, trial: Trial) -> None:
        self.aggregated_parameter_space = {i: [] for i in range(-1, trial.parameter_space.get_dim())}

    def find_target_value(self, target_value: ResultType) -> Mapping | None:
        # find_exact 用
        # NOTE: 並列処理してもよい
        for trial in self.trials:
            finding = trial.find_target_value(target_value)
            if finding:
                return finding
        return None

    def check_timeout_trial(self, now: datetime, timeout_seconds: int) -> list[str]:
        new_trials = []
        outdated_ids = []
        for trial in self.trials:
            if trial.trial_status != TrialStatus.running:
                new_trials.append(trial)
                continue
            delta_sec = trial.measure_seconds_from_registered(now)
            if delta_sec < timeout_seconds:
                # まだ期限内
                new_trials.append(trial)
            else:
                outdated_ids.append(trial.trial_id)
        self.trials = new_trials
        return outdated_ids

    def to_model(self) -> TrialTableModel:
        if self.aggregated_parameter_space is None:
            aps = None
        else:
            aps = {d: [space.to_model() for space in spaces] for d, spaces in self.aggregated_parameter_space.items()}
        return TrialTableModel(
            trials=[trial.to_model() for trial in self.trials],
            aggregated_parameter_space=aps,
        )

    @staticmethod
    def from_model(model: TrialTableModel) -> TrialTable:
        if model.aggregated_parameter_space is None:
            aps = None
        else:
            aps = {
                d: [ParameterAlignedSpace.from_model(space) for space in spaces]
                for d, spaces in model.aggregated_parameter_space.items()
            }
        return TrialTable(
            trials=[Trial.from_model(trial) for trial in model.trials],
            aggregated_parameter_space=aps,
        )



from __future__ import annotations

import json
import logging
from pathlib import Path

from pydantic import BaseModel, Field

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class TableConfig(BaseModel):
    port: int = Field(
        default=8000,
        description="The port number to use table node",
        ge=0,
        le=65535,
    )
    trial_timeout_seconds: int = Field(
        default=600,
        description="Timeout seconds before a trial is reserved and registered",
        ge=1,
    )
    timeout_check_interval_seconds: int = Field(
        default=60,
        description="Interval of time to check timeout trials",
        ge=1,
    )
    curriculum_path: Path = Field(
        default=Path.cwd() / "curriculum.json",
        description="Path to the curriculum json file",
    )
    curriculum_save_interval_seconds: int = Field(
        default=600,
        description="Interval of time to save curriculum json file",
        ge=1,
    )

    @staticmethod
    def load_from_file(path: Path | None) -> TableConfig:
        if path is None:
            path = Path.cwd() / "table_config.json"
        if not path.exists():
            logger.info("TableConfig not found, creating it now")
            default_config = TableConfig()
            with path.open(mode="w") as f:
                json.dump(default_config.model_dump(mode="json"), f, indent=4)
        with path.open() as f:
            return TableConfig.model_validate(json.load(f))


class WorkerConfig(BaseModel):
    name: str | None = Field(
        default=None,
        description="Name of the worker node",
    )
    process_num: int | None = Field(
        default=None,
        description="The number of processes on using `AutoMPTrialRunner`. If `None`, use `os.cpu_count()`.",
    )
    chunk_size: int = Field(
        default=1,
        description="The size of the chunks to be passed to each process.",
    )
    max_size: int = Field(
        default=1,
        description="The maximum size of a trial.",
    )
    disable_function_progress_bar: bool = Field(
        default=False,
        description="Whether to disable progress bar.",
    )
    retaining_capacity: set[str] = Field(
        default_factory=set,
        description="Set of capabilities that the worker node has.",
    )
    wait_seconds_on_no_trial: int = Field(
        default=5,
        description="Waiting time when there was no trial allocated by the table node.",
        ge=1,
    )
    table_node_request_timeout_seconds: int = Field(
        default=30,
        description="Timeout for requests to table nodes.",
        ge=1,
    )


class TableConfigProvider:
    _TABLE: TableConfig | None = None

    @classmethod
    def set(cls, config: TableConfig) -> None:
        cls._TABLE = config

    @classmethod
    def get(cls, table_config_path: Path | None = None) -> TableConfig:
        if cls._TABLE is None:
            cls._TABLE = TableConfig.load_from_file(table_config_path)
        return cls._TABLE
